# pylint: disable=unused-import,missing-docstring

import autofaiss.datasets
import autofaiss.external
import autofaiss.indices
import autofaiss.metrics
import autofaiss.utils
from autofaiss.version import __author__, __version__
